"""
utilities for directories
"""

from pathlib import Path
import tempfile
import shutil
from uuid import uuid4


def create_temp_working_directory() -> str:
    """Create a temporary working directory

    :rtype: str
    :returns: The path to the temporary working directory
    """
    working_directory = Path(tempfile.gettempdir()).joinpath(str(uuid4()))
    working_directory.mkdir(exist_ok=True)

    return working_directory.as_posix()


def delete_temp_working_directory(working_directory: str) -> None:
    """Delete the temporary working directory

    :type working_directory: str
    :param working_directory: The path to the temporary working

    :rtype: None
    :returns: Nothing it deletes the directory
    """
    shutil.rmtree(working_directory)



from pod_porter.pod_porter_cli import cli_argument_parser


def test_cli_argument_parser_run_job_template():
    arg_parser = cli_argument_parser()
    args = arg_parser.parse_args(
        [
            "template",
            "-n",
            "some-release",
            "-m",
            "some-map",
        ]
    )

    assert args.which_sub == "template"
    assert args.name == "some-release"
    assert args.map == "some-map"



"""
CLI for pod_porter
"""

import json
from pathlib import Path
from argparse import ArgumentParser
from importlib.metadata import version

from yaml import safe_load
from pod_porter.pod_porter import PorterMapsRunner, validate_json_against_schema
from pod_porter.util.file_read_write import write_file, create_tar_gz_file, extract_tar_gz_file, create_new_map


def common_sub_parser_map_arguments(sub_arg_parser: ArgumentParser) -> None:
    """Add common arguments to a sub parser

    :type sub_arg_parser: ArgumentParser
    :param sub_arg_parser: The sub parser to add the arguments to

    :rtype: None
    :returns: Nothing it appends the arguments to the sub parser
    """
    sub_arg_parser.add_argument("-m", "--map", required=True, help="Path to the map")


def common_sub_parser_arguments(sub_arg_parser: ArgumentParser) -> None:
    """Add common arguments to a sub parser

    :type sub_arg_parser: ArgumentParser
    :param sub_arg_parser: The sub parser to add the arguments to

    :rtype: None
    :returns: Nothing it appends the arguments to the sub parser
    """
    sub_arg_parser.add_argument("-n", "--name", required=True, help="Release name")
    common_sub_parser_map_arguments(sub_arg_parser=sub_arg_parser)
    sub_arg_parser.add_argument(
        "-f",
        "--file-values",
        required=False,
        default=None,
        help="Path to the values you want to use instead of the map values",
    )


def common_sub_parser_output_path(sub_arg_parser: ArgumentParser) -> None:
    """Add common arguments to a sub parser

    :type sub_arg_parser: ArgumentParser
    :param sub_arg_parser: The sub parser to add the arguments to

    :rtype: None
    :returns: Nothing it appends the arguments to the sub parser
    """
    sub_arg_parser.add_argument("-o", "--output", required=True, help="Path to output file/files")


def cli_argument_parser() -> ArgumentParser:
    """Function to create the argument parser

    :rtype: ArgumentParser
    :returns: The argument parser
    """
    arg_parser = ArgumentParser(description=f"pod-porter version: {version('pod-porter')}")
    subparsers = arg_parser.add_subparsers(
        title="commands",
        description="Valid commands: a single command is required",
        help="CLI Help",
        dest="a single command please see the -h option",
    )
    subparsers.required = True

    # This is the sub parser to print the rendered compose file
    arg_parser_template = subparsers.add_parser("template", help="View the rendered compose file")
    arg_parser_template.set_defaults(which_sub="template")
    common_sub_parser_arguments(sub_arg_parser=arg_parser_template)

    # This is the sub parser to write the rendered compose file
    arg_parser_template_write = subparsers.add_parser("write", help="Write the rendered compose file")
    arg_parser_template_write.set_defaults(which_sub="template_write")
    common_sub_parser_arguments(sub_arg_parser=arg_parser_template_write)
    common_sub_parser_output_path(sub_arg_parser=arg_parser_template_write)

    # This is the sub parser to package the map
    arg_parser_map_package = subparsers.add_parser("package", help="Package the map (tar.gz) the map")
    arg_parser_map_package.set_defaults(which_sub="package")
    common_sub_parser_map_arguments(sub_arg_parser=arg_parser_map_package)
    common_sub_parser_output_path(sub_arg_parser=arg_parser_map_package)

    # This is the sub parser to un-package the map
    arg_parser_map_unpackage = subparsers.add_parser("un-package", help="Un-Package the map extract (tar.gz)")
    arg_parser_map_unpackage.set_defaults(which_sub="un-package")
    common_sub_parser_map_arguments(sub_arg_parser=arg_parser_map_unpackage)
    common_sub_parser_output_path(sub_arg_parser=arg_parser_map_unpackage)

    # This is the sub parser to create a new map
    arg_parser_map_create = subparsers.add_parser("create", help="Create a new map, with some examples")
    arg_parser_map_create.set_defaults(which_sub="create")
    common_sub_parser_map_arguments(sub_arg_parser=arg_parser_map_create)

    # This is the sub parser to validate a compose file
    arg_parser_validate_compose = subparsers.add_parser("validate-compose", help="Validate a compose file")
    arg_parser_validate_compose.set_defaults(which_sub="validate-compose")
    arg_parser_validate_compose.add_argument("-f", "--file", required=True, help="Path to the compose file")

    # This is the sub parser to validate values file against schema
    arg_parser_validate_values = subparsers.add_parser(
        "validate-values", help="Validate a values file against a schema"
    )
    arg_parser_validate_values.set_defaults(which_sub="validate-values")
    arg_parser_validate_values.add_argument("-f", "--file", required=True, help="Path to the values file")
    arg_parser_validate_values.add_argument("-s", "--schema", required=True, help="Path to the values-schema.json")

    return arg_parser


def cli() -> None:  # pragma: no cover # pylint: disable=too-many-branches,too-many-statements
    """Function to run the command line

    :rtype: None
    :returns: Nothing it is the CLI
    """
    arg_parser = None

    try:
        arg_parser = cli_argument_parser()
        args = arg_parser.parse_args()

        if args.which_sub == "template":
            obj = PorterMapsRunner(path=args.map, release_name=args.name, values_override=args.file_values)
            print(obj.render_compose())

        if args.which_sub == "template_write":
            obj = PorterMapsRunner(path=args.map, release_name=args.name, values_override=args.file_values)
            map_name = obj.get_map_data().name
            map_version = obj.get_map_data().version
            file_name = f"{args.name}-{map_name}-{map_version}.yaml"
            write_file(path=args.output, file_name=file_name, data=obj.render_compose())

        if args.which_sub == "package":
            obj = PorterMapsRunner(path=args.map)
            map_name = obj.get_map_data().name
            map_version = obj.get_map_data().version
            file_name = f"{map_name}-{map_version}.tar.gz"
            create_tar_gz_file(path=args.map, file_name=file_name, output_path=args.output)

        if args.which_sub == "un-package":
            extract_tar_gz_file(path=args.map, output_path=args.output)

        if args.which_sub == "create":
            create_new_map(map_name_and_path=args.map)

        if args.which_sub == "validate-compose":
            compose_file_path: Path = Path(args.file)

            if not compose_file_path.is_file():
                raise FileNotFoundError(f"File not found: {compose_file_path}")

            data = safe_load(compose_file_path.read_text(encoding="utf-8"))

            try:
                PorterMapsRunner.validate_compose_json_schema(compose_data=data)

            except ValueError as error:
                print(error)

        if args.which_sub == "validate-values":
            values_file = Path(args.file)

            if not values_file.is_file():
                raise FileNotFoundError(f"File not found: {values_file}")

            values_file_data = safe_load(values_file.read_text(encoding="utf-8"))

            values_schema_file = Path(args.schema)

            if not values_schema_file.is_file():
                raise FileNotFoundError(f"File not found: {values_schema_file}")

            values_schema_data = json.loads(values_schema_file.read_text(encoding="utf-8"))

            try:
                validate_json_against_schema(values_data=values_file_data, json_schema=values_schema_data)

            except ValueError as error:
                print(error)

    except AttributeError as error:
        print(f"\n !!! {error} !!! \n")
        arg_parser.print_help()

    except FileNotFoundError as error:
        print(f"\n !!! {error} !!! \n")
        arg_parser.print_help()

    except FileExistsError as error:
        print(f"\n !!! {error} !!! \n")
        arg_parser.print_help()

    except Exception as error:  # pylint: disable=broad-exception-caught
        print(f"\n !!! {error} !!! \n")
        arg_parser.print_help()



import pytest
from pod_porter.util.schemas import MapSchema


map_schema_table = [
    ("map_data", None),
    ("map_data_bad_api_version", ValueError),
    ("map_data_bad_name", ValueError),
    ("map_data_bad_description", ValueError),
    ("map_data_bad_version", ValueError),
    ("map_data_bad_app_version", ValueError),
]


@pytest.mark.parametrize("data, error", map_schema_table)
def test_map_schema(data, error, request, map_data_as_json, map_data_as_dict, map_data_as_yaml):
    if error:
        with pytest.raises(error):
            MapSchema(**request.getfixturevalue(data))

    else:
        obj = MapSchema(**request.getfixturevalue(data))
        assert obj.dict() == map_data_as_dict
        assert obj.json() == map_data_as_json
        assert obj.yaml() == map_data_as_yaml



import pytest
from pod_porter.pod_porter import PorterMapsRunner


porter_map_runner_table = [
    ("single_map_path", "single_map_rendered_path", "other_values_none"),
    ("single_map_path", "single_map_rendered_path_other_values", "other_values_file"),
    ("multi_map_path", "multi_map_rendered_path", "other_values_none"),
    ("multi_map_path", "multi_map_rendered_path", "other_values_file_with_global"),
    ("multi_map_with_ignore_path", "multi_map_with_ignore_rendered_path", "other_values_none"),
    ("single_single_map_with_validation_path", "single_single_map_with_validation_rendered_path", "other_values_none"),
]


@pytest.mark.parametrize("map_path, map_render, other_values", porter_map_runner_table)
def test_porter_map(map_path, map_render, other_values, request):
    obj = PorterMapsRunner(
        path=request.getfixturevalue(map_path),
        release_name="thing",
        values_override=request.getfixturevalue(other_values),
    )

    assert obj.render_compose() == open(request.getfixturevalue(map_render), "r").read()
    assert str(obj) == f'PorterMapRunner(path="{request.getfixturevalue(map_path)}", release_name="thing")'


porter_map_runner_bad_table = [
    ("map_path_bad_map", TypeError),
    ("map_path_bad_values", FileNotFoundError),
    ("map_path_no_map", FileNotFoundError),
]


@pytest.mark.parametrize("path, error", porter_map_runner_bad_table)
def test_porter_map_runner_bad(path, error, request):
    if error:
        with pytest.raises(error):
            PorterMapsRunner(path=request.getfixturevalue(path), release_name="thing")



import pytest
from pod_porter.render.render import Render


def test_render_from_string():
    obj = Render()
    response = obj.from_string(template_string="Hello, {{ name }}!", render_vars={"name": "World"})

    assert response == "Hello, World!"


def test_render_from_fileg(templates_path):
    obj = Render(templates_dir=templates_path)
    response = obj.from_file(template_name="test-template.j2", render_vars={"data": "Hello, World!"})

    assert response == "Hello, World!"



import os
import pytest
from pod_porter.util.directories import create_temp_working_directory, delete_temp_working_directory


def test_create_temp_working_directory():
    working_directory = create_temp_working_directory()
    assert os.path.exists(working_directory)
    delete_temp_working_directory(working_directory)
    assert not os.path.exists(working_directory)



"""
utility functions for reading and writing files
"""

import os
import tarfile
import fnmatch
from pod_porter.render.render import Render


def write_file(path: str, file_name: str, data: str) -> None:
    """Write data to a file

    :type path: str
    :param path: The path to the file
    :type file_name: str
    :param file_name: The name of the file
    :type data: str
    :param data: The data to write to the file

    :rtype: None
    :returns: Nothing it writes the data to the file
    """
    with open(os.path.join(path, file_name), "w", encoding="utf-8", newline="\n") as file:
        file.write(data)


def get_ignore_patterns(ignore_file_path: str) -> list:
    """Get the list of files and directories to ignore

    :type ignore_file_path: str
    :param ignore_file_path: The path to the ignore file

    :rtype: list
    :returns: A list of files and directories to ignore
    """
    ignore_list = []
    if not os.path.isfile(ignore_file_path):
        return ignore_list

    with open(ignore_file_path, "r", encoding="utf-8") as file:
        for line in file:
            if line.startswith("#"):
                continue

            if line.strip() == "":
                continue

            ignore_list.append(line.strip())

    return ignore_list


def should_ignore(path: str, patterns: list) -> bool:
    """Check if a file or directory should be ignored

    :type path: str
    :param path: The full path to the file or directory
    :type patterns: list
    :param patterns: The list of files and directories to ignore

    :rtype: bool
    :returns: True if the file or directory should be ignored, False otherwise
    """
    for pattern in patterns:
        if pattern.startswith("!"):
            if path.startswith(pattern[1:]):
                return False

        if fnmatch.fnmatch(path, pattern):
            return True

        if fnmatch.fnmatch(os.path.basename(path), pattern):
            return True

    return False


def create_new_map(map_name_and_path: str) -> None:
    """Create a new map

    :type map_name_and_path: str
    :param map_name_and_path: The full path to create the new map

    :rtype: None
    :returns: Nothing it creates the new map
    """
    os.makedirs(map_name_and_path)
    map_templates_path = os.path.join(map_name_and_path, "templates")
    os.mkdir(map_templates_path)
    render_vars = {"data": {"map_name": os.path.split(map_name_and_path)[1]}}

    render = Render()

    map_file = render.from_file(template_name="new-map.j2", render_vars=render_vars)
    values_file = render.from_file(template_name="new-values.j2", render_vars=render_vars)
    ignore_file = render.from_file(template_name="new-porterignore.j2", render_vars=render_vars)
    service_file = render.from_file(template_name="new-service.j2", render_vars=render_vars)
    volumes_file = render.from_file(template_name="new-volumes.j2", render_vars=render_vars)

    write_file(path=map_name_and_path, file_name="Map.yaml", data=map_file)
    write_file(path=map_name_and_path, file_name="values.yaml", data=values_file)
    write_file(path=map_name_and_path, file_name=".porterignore", data=ignore_file)
    write_file(path=map_templates_path, file_name="service-example.yaml", data=service_file)
    write_file(path=map_templates_path, file_name="volumes-example.yaml", data=volumes_file)


def create_tar_gz_file(path: str, file_name: str, output_path: str) -> None:
    """Create a tar.gz file

    :type path: str
    :param path: The path to the directory to tar.gz
    :type file_name: str
    :param file_name: The name of the file
    :type output_path: str
    :param output_path: The path to save the tar.gz file

    :rtype: None
    :returns: Nothing it creates the tar.gz file
    """
    exclude_patters = get_ignore_patterns(os.path.join(path, ".porterignore"))
    with tarfile.open(os.path.join(output_path, file_name), "w:gz") as tar:
        for dir_path, dir_names, file_names in os.walk(path):
            for pattern in exclude_patters:
                if pattern.endswith("/"):
                    if os.path.normpath(pattern) in dir_names:
                        dir_names.remove(os.path.normpath(pattern))
                    continue

            for file in file_names:
                file_path = os.path.join(dir_path, file)

                if should_ignore(file_path, exclude_patters):
                    continue

                tar.add(file_path, arcname=os.path.join(os.path.basename(path), os.path.relpath(file_path, path)))


def extract_tar_gz_file(path: str, output_path: str) -> None:
    """Extract a tar.gz file

    :type path: str
    :param path: The full path to the tar.gz
    :type output_path: str
    :param output_path: The path to save the extracted files

    :rtype: None
    :returns: Nothing it extracts the tar.gz file
    """
    with tarfile.open(path, "r:gz") as tar:
        tar.extractall(path=output_path)



"""
Specific Data Schemas for the Pod Porter Application
"""

from typing import Literal
import json
import re
from dataclasses import dataclass

import yaml


@dataclass
class _BaseSchema:
    """Base Dataclass Schema"""

    def dict(self) -> dict:
        """Get a dict representation of the Dataclass

        :rtype: dict
        :returns: A dict representation of the Dataclass
        """
        return self.__dict__

    def json(self, pretty=False) -> str:
        """Get a JSON representation of the Dataclass

        :type pretty: bool
        :param pretty: Get pretty JSON

        :rtype: str
        :returns: A JSON representation of the Dataclass
        """
        if pretty:
            indent = 4

        else:
            indent = None

        return json.dumps(self.dict(), indent=indent)

    def yaml(self) -> str:
        """Get a YAML representation of the Dataclass

        :rtype: str
        :returns: A YAML representation of the Dataclass
        """
        return yaml.safe_dump(self.dict(), indent=2)

    @staticmethod
    def validate_string(name: str, value: str) -> None:
        """Validate a string

        :type value: str
        :param value: Value to validate
        :type name: str
        :param name: Name of the value
        """
        if not isinstance(value, str):
            raise TypeError(f"{name} must be a string, but received {type(value)}")

    @staticmethod
    def validate_string_no_spaces(name: str, value: str) -> None:
        """Validate a string to have no spaces

        :type value: str
        :param value: Value to validate
        :type name: str
        :param name: Name of the value
        """
        regex = re.compile(r"^\S+$")

        if not regex.match(value):
            raise ValueError(f"{name} must not contain spaces")

    @staticmethod
    def validate_string_no_spaces_begin_end(name: str, value: str) -> None:
        """Validate a string to have no spaces at the beginning or end

        :type value: str
        :param value: Value to validate
        :type name: str
        :param name: Name of the value
        """
        regex = re.compile(r"^\S.*\S$")

        if not regex.match(value):
            raise ValueError(f"{name} must not contain spaces at the beginning or end")

    @staticmethod
    def validate_api_version(name: str, value: str) -> None:
        """Validate the api version

        :type value: str
        :param value: Value to validate
        :type name: str
        :param name: Name of the value
        """
        allowed_versions = ["v1"]

        if value not in allowed_versions:
            raise ValueError(f"{name} must be one of {allowed_versions}")

    @staticmethod
    def validate_semantic_version(name: str, value: str) -> None:
        """Validate a string is a semantic version

        :type value: str
        :param value: Value to validate
        :type name: str
        :param name: Name of the value
        """
        regex = re.compile(r"^\d+\.\d+\.\d+$")

        if not regex.match(value):
            raise ValueError(f"{name} must be a semantic version")


@dataclass
class MapSchema(_BaseSchema):
    """Map.yaml Schema

    :type api_version: Literal["v1"]
    :cvar api_version: Name of the Inventory
    :type name: str
    :cvar name: Name of the Inventory
    :type description: str
    :cvar description: Description of the Inventory
    :type version: str
    :cvar version: Description of the Inventory
    :type app_version: str
    :cvar app_version: Description of the Inventory
    """

    api_version: Literal["v1"]
    name: str
    description: str
    version: str
    app_version: str

    def __post_init__(self):
        self.validate_api_version("api_version", self.api_version)
        self.validate_string_no_spaces("name", self.name)
        self.validate_string_no_spaces_begin_end("description", self.description)
        self.validate_semantic_version("version", self.version)
        self.validate_string_no_spaces("app_version", self.app_version)



import os
import shutil
import pytest
from pod_porter.util.file_read_write import write_file, create_tar_gz_file, extract_tar_gz_file, create_new_map


def test_write_file(tmp_path):
    path = tmp_path

    write_file(path=str(path), file_name="test.txt", data="Hello, World!")

    assert os.path.exists(path / "test.txt")


create_tar_gz_file_table = [
    ("single_map_path"),
    ("multi_map_path"),
    ("multi_map_with_ignore_path"),
]


@pytest.mark.parametrize("map_type", create_tar_gz_file_table)
def test_create_tar_gz_file(tmp_path, map_type, request):
    the_temp_path_path = str(tmp_path)
    tar_file_name = "test.tar.gz"
    dst_path = os.path.join(the_temp_path_path, "mongo")

    shutil.copytree(request.getfixturevalue(map_type), dst_path)

    create_tar_gz_file(
        path=os.path.join(the_temp_path_path, "mongo"), file_name=tar_file_name, output_path=the_temp_path_path
    )

    assert os.path.exists(os.path.join(the_temp_path_path, tar_file_name))

    shutil.rmtree(dst_path)

    extract_tar_gz_file(path=os.path.join(the_temp_path_path, tar_file_name), output_path=the_temp_path_path)

    assert os.path.exists(os.path.join(the_temp_path_path, dst_path))


def test_create_new_map(tmp_path):
    the_temp_path_path = str(tmp_path)
    map_name = "test-map"

    create_new_map(map_name_and_path=os.path.join(the_temp_path_path, map_name))

    assert os.path.exists(os.path.join(the_temp_path_path, map_name))



"""
Pod Porter Main Application
"""

from typing import List, Optional, Dict
from pathlib import Path
import json
import os
from yaml import safe_load, safe_dump
from jsonschema import validate, Draft202012Validator, ValidationError
from pod_porter.render.render import Render
from pod_porter.util.directories import create_temp_working_directory, delete_temp_working_directory
from pod_porter.util.file_read_write import write_file, extract_tar_gz_file
from pod_porter.util.schemas import MapSchema

JSON_SCHEMA_FORMAT_CHECKERS = Draft202012Validator.FORMAT_CHECKER
COMPOSE_SPEC: Path = Path(__file__).parent.joinpath("render").joinpath("compose_spec").joinpath("compose-spec.json")


def validate_json_against_schema(values_data: dict, json_schema: dict) -> None:
    """Validate a data set against a JSON schema

    :type values_data: dict
    :param values_data: The data to validate
    :type json_schema: dict
    :param json_schema: The schema to validate against

    :rtype: None
    :returns: Nothing

    :raises: ValueError: If the validation fails
    """
    try:
        validate(instance=values_data, schema=json_schema, format_checker=Draft202012Validator.FORMAT_CHECKER)

    except ValidationError as err:
        readable_error = (
            f"Error in validating compose file!\nArgument Error: {err.args}\n"
            f"In Schema Path: {'.'.join(err.absolute_path)}"
        )
        raise ValueError(readable_error) from err


class _PorterMap:  # pylint: disable=too-many-instance-attributes
    """A class to represent the PorterMap

    :type path: str
    :param path: The path to the directory containing the map.yaml and values.yaml files
    :type release_name: Optional[str] = None
    :param release_name: The name of the release
    :type values_override: values_override: Optional[str] = None
    :param values_override: The path to the yaml to override with
    :type top_level: bool = True
    :param top_level: If the map is a top level map
    :type top_level_path: Optional[str] = None
    :param top_level_path: The path to the top level map

    :rtype: None
    :returns: Nothing
    """

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        path: str,
        release_name: Optional[str] = None,
        values_override: Optional[str] = None,
        top_level: bool = True,
        top_level_path: Optional[str] = None,
    ) -> None:
        self._name = os.path.basename(path)
        self._top_level = top_level
        self._top_level_path = os.path.abspath(top_level_path)
        self._temp_working_directory = create_temp_working_directory()
        self._path = os.path.abspath(path)
        self._release_name = release_name or "release-name"
        self._map_data = self._get_map()
        self._values_data = self._get_values(values_override=values_override)

        if not self._map_data:
            raise ValueError("map_data is empty")

        self._templates = self._get_templates(templates_path=os.path.join(self._path, "templates"))
        self._pre_render()
        self._templates_pre_render = self._get_templates(templates_path=self._temp_working_directory)
        self._compose = {}
        self._services = self._get_service_templates()
        self._configs = self._get_config_templates()
        self._volumes = self._get_volume_templates()
        self._secrets = self._get_secrets_templates()
        self._networks = self._get_network_templates()

    def __repr__(self) -> str:
        """Return the string representation of the object

        :rtype: str
        :returns: The string representation of the object
        """
        return f'PorterMap(path="{self._path}", release_name="{self._release_name}")'

    @staticmethod
    def validate_value_json_schema(values_data: dict, values_path: str) -> None:
        """Validate data against a JSON schema.

        :type values_data: dict
        :param values_data: The values data to validate
        :type values_path: str
        :param values_path: The path to the values file

        :rtype: None
        :returns: Nothing it validates the JSON data against the JSON schema
        """
        json_schema_path = os.path.join(os.path.split(values_path)[0], "values-schema.json")

        if not os.path.isfile(json_schema_path):
            return

        with open(json_schema_path, "r", encoding="utf-8") as json_schema_file:
            json_schema = json.load(json_schema_file)

        validate_json_against_schema(values_data=values_data, json_schema=json_schema)

    def get_services(self) -> dict:
        """Get the services data

        :rtype: dict
        :returns: The services data
        """
        return self._services

    def get_configs(self) -> dict:
        """Get the configs data

        :rtype: dict
        :returns: The configs data
        """
        return self._configs

    def get_volumes(self) -> dict:
        """Get the volumes data

        :rtype: dict
        :returns: The volumes data
        """
        return self._volumes

    def get_secrets(self) -> dict:
        """Get the secrets data

        :rtype: dict
        :returns: The secrets data
        """
        return self._secrets

    def get_networks(self) -> dict:
        """Get the networks data

        :rtype: dict
        :returns: The networks data
        """
        return self._networks

    def get_temp_working_directory(self) -> str:
        """Get the temp working directory

        :rtype: str
        :returns: The working directory
        """
        return self._temp_working_directory

    def get_map_data(self) -> MapSchema:
        """Get the map data

        :rtype: MapSchema
        :returns: The map data
        """
        return self._map_data

    @staticmethod
    def get_yaml_data(path: str) -> dict:
        """Load the data from a yaml file and return it

        :type path: str
        :param path: The path to the yaml file

        :rtype: dict
        :returns: The data from the yaml
        """
        with open(path, "r", encoding="utf-8") as file:
            data = safe_load(file.read())

        return data

    @staticmethod
    def _get_templates(templates_path: str) -> List[str]:
        """Get a list of all the template files in the templates directory

        :type templates_path: str
        :param templates_path: The path to the templates directory

        :rtype: List[str]
        :returns: A list of all the template files in the templates directory
        """
        template_files = []

        for item in os.listdir(templates_path):
            if os.path.isfile(os.path.join(templates_path, item)):
                template_files.append(os.path.abspath(os.path.join(templates_path, item)))

        return template_files

    def _get_compose_type_data(self, compose_type: str) -> dict:
        """Get the data for a specific compose type from the templates

        :type compose_type: str
        :param compose_type: The compose type to get the data for

        :rtype: dict
        :returns: The data for the compose type
        """
        services = {compose_type: {}}
        for template in self._templates_pre_render:
            template_dict = self.get_yaml_data(template)

            if not template_dict:
                continue

            if template_dict.get(compose_type):
                services.get(compose_type).update(template_dict[compose_type])

        return services

    def _get_service_templates(self) -> dict:
        """Get the service data from the templates

        :rtype: dict
        :returns: The service data from the templates
        """
        services = self._get_compose_type_data("services")

        self._compose.update(services)

        return services

    def _get_volume_templates(self) -> dict:
        """Get the volume data from the templates

        :rtype: dict
        :returns: The volume data from the templates
        """
        volumes = self._get_compose_type_data("volumes")

        self._compose.update(volumes)

        return volumes

    def _get_network_templates(self) -> dict:
        """Get the network data from the templates

        :rtype: dict
        :returns: The network data from the templates
        """
        networks = self._get_compose_type_data("networks")

        self._compose.update(networks)

        return networks

    def _get_config_templates(self) -> dict:
        """Get the config data from the templates

        :rtype: dict
        :returns: The config data from the templates
        """
        configs = self._get_compose_type_data("configs")

        self._compose.update(configs)

        return configs

    def _get_secrets_templates(self) -> dict:
        """Get the secrets data from the templates

        :rtype: dict
        :returns: The secrets data from the templates
        """
        secrets = self._get_compose_type_data("secrets")

        self._compose.update(secrets)

        return secrets

    def _get_map(self) -> MapSchema:
        """Load the map.yaml file and return the data

        :rtype: MapSchema
        :returns: The validated data from the map.yaml
        """
        map_path = os.path.join(self._path, "Map.yaml")

        if not os.path.isfile(map_path):
            raise FileNotFoundError("Map.yaml not found")

        return MapSchema(**self.get_yaml_data(map_path))

    def _get_values(self, values_override: Optional[str] = None) -> dict:
        """Load the values.yaml file and return the data

        :rtype: dict
        :returns: The data from the values.yaml
        """
        if not values_override:
            values_path = os.path.join(self._top_level_path, "values.yaml")

        else:
            values_path = values_override

        if not os.path.isfile(values_path):
            raise FileNotFoundError(f"values file '{values_path}' not found")

        initial_values = self.get_yaml_data(values_path)

        if self._top_level:
            values = initial_values
            self.validate_value_json_schema(values_data=values, values_path=values_path)

        elif not self._top_level and not initial_values.get("sub_map_values"):
            values_path = os.path.join(self._path, "values.yaml")
            values = self.get_yaml_data(values_path)
            self.validate_value_json_schema(values_data=values, values_path=values_path)

        elif not self._top_level and not initial_values.get("sub_map_values").get(self._name):
            values_path = os.path.join(self._path, "values.yaml")
            values = self.get_yaml_data(values_path)
            self.validate_value_json_schema(values_data=values, values_path=values_path)

        else:
            values = self.get_yaml_data(values_path).get("sub_map_values").get(self._name)

        return {"values": values, "release": {"name": self._release_name}}

    def _pre_render(self) -> None:
        """Pre-render the templates from the map

        :rtype: None
        :returns: Nothing it writes rendered templates to the temp working directory
        """
        templates_path = os.path.join(self._path, "templates")
        render_obj = Render(templates_dir=templates_path)
        for path in self._templates:
            template = os.path.split(path)[1]
            write_file(
                self._temp_working_directory,
                template,
                render_obj.from_file(template_name=template, render_vars=self._values_data),
            )


class PorterMapsRunner:  # pylint: disable=too-many-instance-attributes
    """A class to represent the PorterMapRunner for collecting and running maps

    :type path: str
    :param path: The path to the directory containing the map.yaml and values.yaml files
    :type release_name: Optional[str] = None
    :param release_name: The name of the release

    :rtype: None
    :returns: Nothing
    """

    def __init__(self, path: str, release_name: Optional[str] = None, values_override: Optional[str] = None) -> None:
        self._path = path
        self._release_name = release_name or "release-name"
        if values_override:
            self._values_override = os.path.abspath(values_override)

        else:
            self._values_override = None

        self._toplevel_map_data = None
        self._all_maps = self._collect_maps()
        self._services = {"services": {}}
        self._configs = {"configs": {}}
        self._volumes = {"volumes": {}}
        self._secrets = {"secrets": {}}
        self._networks = {"networks": {}}
        self._compose = {}
        self._merge_maps()

    def __repr__(self) -> str:
        """Return the string representation of the object

        :rtype: str
        :returns: The string representation of the object
        """
        return f'PorterMapRunner(path="{self._path}", release_name="{self._release_name}")'

    def get_map_data(self) -> MapSchema:
        """Get the map data

        :rtype: MapSchema
        :returns: The map data
        """
        return self._toplevel_map_data

    def _collect_maps(self) -> List[Dict]:
        """Collect all the maps in the directory

        :rtype: List[Dict]
        :returns: A list of PorterMap objects
        """
        top_level_map = _PorterMap(
            path=self._path,
            release_name=self._release_name,
            values_override=self._values_override,
            top_level_path=self._path,
        )

        self._toplevel_map_data = top_level_map.get_map_data()

        maps = [{"map_obj": top_level_map, "map_name": os.path.basename(self._path)}]

        if os.path.isdir(os.path.join(self._path, "maps")):
            for single_map in os.listdir(os.path.join(self._path, "maps")):
                if single_map.endswith(".tar.gz"):
                    tar_path = os.path.join(self._path, "maps", single_map)
                    extract_path = os.path.join(self._path, "maps")
                    extract_tar_gz_file(path=tar_path, output_path=extract_path)

            for single_map in os.listdir(os.path.join(self._path, "maps")):
                if single_map.endswith(".tar.gz"):
                    continue

                maps.append(
                    {
                        "map_obj": _PorterMap(
                            path=os.path.join(self._path, "maps", single_map),
                            release_name=self._release_name,
                            values_override=self._values_override,
                            top_level=False,
                            top_level_path=self._path,
                        ),
                        "map_name": single_map,
                    }
                )

        return maps

    @staticmethod
    def validate_compose_json_schema(compose_data: dict) -> None:
        """Validate the created compose data against the compose spec JSON schema
        https://github.com/compose-spec/compose-spec/blob/main/schema/compose-spec.json

        :type compose_data: dict
        :param compose_data: The compose data

        :rtype: None
        :returns: Nothing it validates the JSON data against the JSON schema
        """
        if not COMPOSE_SPEC.is_file():
            raise FileNotFoundError(f"compose file json spec not found {COMPOSE_SPEC.as_posix()}")

        with open(COMPOSE_SPEC.as_posix(), "r", encoding="utf-8") as compose_schema_file:
            compose_json_schema = json.load(compose_schema_file)

        validate_json_against_schema(values_data=compose_data, json_schema=compose_json_schema)

    def render_compose(self) -> str:
        """Render the compose file

        :rtype: str
        :returns: The rendered compose file
        """
        render_obj = Render()

        compose_rendered = render_obj.from_file(
            template_name="compose-layout.j2", render_vars={"compose_data": safe_dump(self._compose)}
        )

        self.validate_compose_json_schema(compose_data=safe_load(compose_rendered))

        return compose_rendered

    def _merge_maps(self) -> None:
        """Merge all the maps into a single compose file

        :rtype: None
        :returns: Nothing
        """
        for single_map in self._all_maps:
            self._services["services"].update(single_map.get("map_obj").get_services().get("services"))
            self._configs["configs"].update(single_map.get("map_obj").get_configs().get("configs"))
            self._volumes["volumes"].update(single_map.get("map_obj").get_volumes().get("volumes"))
            self._secrets["secrets"].update(single_map.get("map_obj").get_secrets().get("secrets"))
            self._networks["networks"].update(single_map.get("map_obj").get_networks().get("networks"))
            delete_temp_working_directory(single_map.get("map_obj").get_temp_working_directory())
        self._compose.update(self._services)
        self._compose.update(self._configs)
        self._compose.update(self._volumes)
        self._compose.update(self._secrets)
        self._compose.update(self._networks)



"""
Rendering engine
"""

from typing import Optional
import os.path

from jinja2 import Environment, FileSystemLoader, BaseLoader, select_autoescape


class Render:
    """Render the templates class

    :type templates_dir: Optional[str] = None
    :param templates_dir: The directory where the templates are located

    :rtype: None
    :returns: Nothing
    """

    BASE_DIRECTORY = os.path.abspath(os.path.dirname(__file__))
    TEMPLATE_DIRECTORY = os.path.join(BASE_DIRECTORY, "templates")

    def __init__(self, templates_dir: Optional[str] = None) -> None:
        if templates_dir:
            self.templates_dir = templates_dir

        else:
            self.templates_dir = self.TEMPLATE_DIRECTORY

        self.env_file_render = self.__set_template_file_render_environment()
        self.env_string_render = self.__set_template_string_render_environment()

    def __set_template_file_render_environment(self) -> Environment:
        """Private method to set the template file render environment

        :rtype: jinja2.Environment
        :returns: The template file render environment
        """
        return Environment(
            autoescape=select_autoescape(enabled_extensions="yaml", default_for_string=True),
            loader=FileSystemLoader(searchpath=self.templates_dir),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    @staticmethod
    def __set_template_string_render_environment() -> Environment:
        """Private method to set the template string render environment

        :rtype: jinja2.Environment
        :returns: The template file render environment
        """
        return Environment(
            autoescape=select_autoescape(enabled_extensions="yaml", default_for_string=True),
            loader=BaseLoader(),
            trim_blocks=True,
            lstrip_blocks=True,
        )

    def from_file(self, template_name: str, render_vars: dict) -> str:
        """Render the template from a file

        :type template_name: str
        :param template_name: The name of the template to render
        :type render_vars: dict
        :param render_vars: The variables to render the template with

        :rtype: str
        :returns: The rendered template
        """
        template = self.env_file_render.get_template(name=template_name)
        return template.render(render_vars)

    def from_string(self, template_string: str, render_vars: dict):
        """Render the template from a string

        :type template_string: str
        :param template_string: The string of the template to render
        :type render_vars: dict
        :param render_vars: The variables to render the template with

        :rtype: str
        :returns: The rendered template
        """
        template = self.env_string_render.from_string(source=template_string)
        return template.render(render_vars)
